@@ -18,6 +18,7 @@ package controller
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"time"
 
@@ -28,6 +29,7 @@ import (
 	storage "k8s.io/api/storage/v1beta1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
+	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/kubernetes"
 	corelisters "k8s.io/client-go/listers/core/v1"
 	storagelisters "k8s.io/client-go/listers/storage/v1beta1"
@@ -190,9 +192,8 @@ func (h *csiHandler) prepareVANodeID(va *storage.VolumeAttachment, nodeID string
 	return clone, true
 }
 
-func (h *csiHandler) saveVA(va *storage.VolumeAttachment) (*storage.VolumeAttachment, error) {
-	// TODO: use patch to save us from VersionError
-	newVA, err := h.client.StorageV1beta1().VolumeAttachments().Update(va)
+func (h *csiHandler) saveVA(va *storage.VolumeAttachment, patch []byte) (*storage.VolumeAttachment, error) {
+	newVA, err := h.client.StorageV1beta1().VolumeAttachments().Patch(va.Name, types.MergePatchType, patch)
 	if err != nil {
 		return va, err
 	}
@@ -214,11 +215,12 @@ func (h *csiHandler) addPVFinalizer(pv *v1.PersistentVolume) (*v1.PersistentVolu
 	klog.V(4).Infof("Adding finalizer to PV %q", pv.Name)
 	clone := pv.DeepCopy()
 	clone.Finalizers = append(clone.Finalizers, finalizerName)
-	// TODO: use patch to save us from VersionError
-	newPV, err := h.client.CoreV1().PersistentVolumes().Update(clone)
+
+	newPV, err := h.patchPV(pv, clone)
 	if err != nil {
 		return pv, err
 	}
+
 	klog.V(4).Infof("PV finalizer added to %q", pv.Name)
 	return newPV, nil
 }
@@ -239,12 +241,6 @@ func getCSISource(pv *v1.PersistentVolume) (*v1.CSIPersistentVolumeSource, error
 	}
 	if pv.Spec.CSI != nil {
 		return pv.Spec.CSI, nil
-	} else if csitranslationlib.IsPVMigratable(pv) {
-		csiPV, err := csitranslationlib.TranslateInTreePVToCSI(pv)
-		if err != nil {
-			return nil, fmt.Errorf("failed to translate in tree pv to CSI: %v", err)
-		}
-		return csiPV.Spec.CSI, nil
 	}
 	return nil, fmt.Errorf("pv contained non-csi source that was not migrated")
 }
@@ -254,26 +250,50 @@ func (h *csiHandler) csiAttach(va *storage.VolumeAttachment) (*storage.VolumeAtt
 	// Check as much as possible before adding VA finalizer - it would block
 	// deletion of VA on error.
 
-	if va.Spec.Source.PersistentVolumeName == nil {
-		return va, nil, fmt.Errorf("VolumeAttachment.spec.persistentVolumeName is empty")
-	}
+	var csiSource *v1.CSIPersistentVolumeSource
+	var pvSpec *v1.PersistentVolumeSpec
+	if va.Spec.Source.PersistentVolumeName != nil {
+		if va.Spec.Source.InlineVolumeSpec != nil {
+			return va, nil, errors.New("both InlineCSIVolumeSource and PersistentVolumeName specified in VA source")
+		}
+		pv, err := h.pvLister.Get(*va.Spec.Source.PersistentVolumeName)
+		if err != nil {
+			return va, nil, err
+		}
+		// Refuse to attach volumes that are marked for deletion.
+		if pv.DeletionTimestamp != nil {
+			return va, nil, fmt.Errorf("PersistentVolume %q is marked for deletion", pv.Name)
+		}
+		pv, err = h.addPVFinalizer(pv)
+		if err != nil {
+			return va, nil, fmt.Errorf("could not add PersistentVolume finalizer: %s", err)
+		}
 
-	pv, err := h.pvLister.Get(*va.Spec.Source.PersistentVolumeName)
-	if err != nil {
-		return va, nil, err
-	}
-	// Refuse to attach volumes that are marked for deletion.
-	if pv.DeletionTimestamp != nil {
-		return va, nil, fmt.Errorf("PersistentVolume %q is marked for deletion", pv.Name)
-	}
-	pv, err = h.addPVFinalizer(pv)
-	if err != nil {
-		return va, nil, fmt.Errorf("could not add PersistentVolume finalizer: %s", err)
-	}
+		if csitranslationlib.IsPVMigratable(pv) {
+			pv, err = csitranslationlib.TranslateInTreePVToCSI(pv)
+			if err != nil {
+				return va, nil, fmt.Errorf("failed to translate in tree pv to CSI: %v", err)
+			}
+		}
 
-	csiSource, err := getCSISource(pv)
-	if err != nil {
-		return va, nil, err
+		// Both csiSource and pvSpec could be translated here if the PV was
+		// migrated
+		csiSource, err = getCSISource(pv)
+		if err != nil {
+			return va, nil, err
+		}
+
+		pvSpec = &pv.Spec
+	} else if va.Spec.Source.InlineVolumeSpec != nil {
+		if va.Spec.Source.InlineVolumeSpec.CSI != nil {
+			csiSource = va.Spec.Source.InlineVolumeSpec.CSI
+		} else {
+			return va, nil, errors.New("inline volume spec contains nil CSI source")
+		}
+
+		pvSpec = va.Spec.Source.InlineVolumeSpec
+	} else {
+		return va, nil, errors.New("neither InlineCSIVolumeSource nor PersistentVolumeName specified in VA source")
 	}
 
 	attributes, err := GetVolumeAttributes(csiSource)
@@ -291,7 +311,7 @@ func (h *csiHandler) csiAttach(va *storage.VolumeAttachment) (*storage.VolumeAtt
 		readOnly = false
 	}
 
-	volumeCapabilities, err := GetVolumeCapabilities(pv, csiSource)
+	volumeCapabilities, err := GetVolumeCapabilities(pvSpec)
 	if err != nil {
 		return va, nil, err
 	}
@@ -308,10 +328,9 @@ func (h *csiHandler) csiAttach(va *storage.VolumeAttachment) (*storage.VolumeAtt
 	originalVA := va
 	va, finalizerAdded := h.prepareVAFinalizer(va)
 	va, nodeIDAdded := h.prepareVANodeID(va, nodeID)
+
 	if finalizerAdded || nodeIDAdded {
-		va, err = h.saveVA(va)
-		if err != nil {
-			// va modification failed, return the original va that's still on API server
+		if va, err = h.patchVA(originalVA, va); err != nil {
 			return originalVA, nil, fmt.Errorf("could not save VolumeAttachment: %s", err)
 		}
 	}
@@ -329,18 +348,33 @@ func (h *csiHandler) csiAttach(va *storage.VolumeAttachment) (*storage.VolumeAtt
 }
 
 func (h *csiHandler) csiDetach(va *storage.VolumeAttachment) (*storage.VolumeAttachment, error) {
-	if va.Spec.Source.PersistentVolumeName == nil {
-		return va, fmt.Errorf("VolumeAttachment.spec.persistentVolumeName is empty")
-	}
-
-	pv, err := h.pvLister.Get(*va.Spec.Source.PersistentVolumeName)
-	if err != nil {
-		return va, err
-	}
-
-	csiSource, err := getCSISource(pv)
-	if err != nil {
-		return va, err
+	var csiSource *v1.CSIPersistentVolumeSource
+	if va.Spec.Source.PersistentVolumeName != nil {
+		if va.Spec.Source.InlineVolumeSpec != nil {
+			return va, errors.New("both InlineCSIVolumeSource and PersistentVolumeName specified in VA source")
+		}
+		pv, err := h.pvLister.Get(*va.Spec.Source.PersistentVolumeName)
+		if err != nil {
+			return va, err
+		}
+		if csitranslationlib.IsPVMigratable(pv) {
+			pv, err = csitranslationlib.TranslateInTreePVToCSI(pv)
+			if err != nil {
+				return va, fmt.Errorf("failed to translate in tree pv to CSI: %v", err)
+			}
+		}
+		csiSource, err = getCSISource(pv)
+		if err != nil {
+			return va, err
+		}
+	} else if va.Spec.Source.InlineVolumeSpec != nil {
+		if va.Spec.Source.InlineVolumeSpec.CSI != nil {
+			csiSource = va.Spec.Source.InlineVolumeSpec.CSI
+		} else {
+			return va, errors.New("inline volume spec contains nil CSI source")
+		}
+	} else {
+		return va, errors.New("neither InlineCSIVolumeSource nor PersistentVolumeName specified in VA source")
 	}
 
 	volumeHandle, _, err := GetVolumeHandle(csiSource)
@@ -381,8 +415,9 @@ func (h *csiHandler) saveAttachError(va *storage.VolumeAttachment, err error) (*
 		Message: err.Error(),
 		Time:    metav1.Now(),
 	}
-	newVa, err := h.client.StorageV1beta1().VolumeAttachments().Update(clone)
-	if err != nil {
+
+	var newVa *storage.VolumeAttachment
+	if newVa, err = h.patchVA(va, clone); err != nil {
 		return va, err
 	}
 	klog.V(4).Infof("Saved attach error to %q", va.Name)
@@ -396,8 +431,9 @@ func (h *csiHandler) saveDetachError(va *storage.VolumeAttachment, err error) (*
 		Message: err.Error(),
 		Time:    metav1.Now(),
 	}
-	newVa, err := h.client.StorageV1beta1().VolumeAttachments().Update(clone)
-	if err != nil {
+
+	var newVa *storage.VolumeAttachment
+	if newVa, err = h.patchVA(va, clone); err != nil {
 		return va, err
 	}
 	klog.V(4).Infof("Saved detach error to %q", va.Name)
@@ -462,12 +498,13 @@ func (h *csiHandler) SyncNewOrUpdatedPersistentVolume(pv *v1.PersistentVolume) {
 		newFinalizers = nil
 	}
 	clone.Finalizers = newFinalizers
-	_, err = h.client.CoreV1().PersistentVolumes().Update(clone)
-	if err != nil {
+
+	if _, err = h.patchPV(pv, clone); err != nil {
 		klog.Errorf("Failed to remove finalizer from PV %q: %s", pv.Name, err.Error())
 		h.pvQueue.AddRateLimited(pv.Name)
 		return
 	}
+
 	klog.V(2).Infof("Removed finalizer from PV %q", pv.Name)
 	h.pvQueue.Forget(pv.Name)
 
@@ -530,3 +567,29 @@ func (h *csiHandler) getNodeID(driver string, nodeName string, va *storage.Volum
 	// return nodeLister.Get error
 	return "", err
 }
+
+func (h *csiHandler) patchVA(va, clone *storage.VolumeAttachment) (*storage.VolumeAttachment, error) {
+	patch, err := createMergePatch(va, clone)
+	if err != nil {
+		return va, err
+	}
+
+	newVa, err := h.client.StorageV1beta1().VolumeAttachments().Patch(va.Name, types.MergePatchType, patch)
+	if err != nil {
+		return va, err
+	}
+	return newVa, nil
+}
+
+func (h *csiHandler) patchPV(pv, clone *v1.PersistentVolume) (*v1.PersistentVolume, error) {
+	patch, err := createMergePatch(pv, clone)
+	if err != nil {
+		return pv, err
+	}
+
+	newPV, err := h.client.CoreV1().PersistentVolumes().Patch(pv.Name, types.MergePatchType, patch)
+	if err != nil {
+		return pv, err
+	}
+	return newPV, nil
+}
